/* ==================================================================
   File:        main.cpp
   Author:      Steve Monks after Chun-Yu
   Description:
	WinAmp Kernel Streaming Output Plugin

 	$Id: out_ks.cpp,v 1.9 2005/12/31 15:19:16 Steve Exp $
	$Log: out_ks.cpp,v $
	Revision 1.9  2005/12/31 15:19:16  Steve
	Corrected revision number to 2.4
	
	Revision 1.8  2005/12/31 15:05:46  Steve
	Fixed bug where config settings disappear if no tracks are playing when
	"release when stopped" is enabled
	
	Revision 1.7  2005/12/27 01:20:01  Steve
	Added Oversampling for <48KHz 16 bit audio
	Significant re-org and clean up of code
	
	Revision 1.6  2005/12/24 16:24:55  Steve
	Final changes for V2.3
	
	Revision 1.5  2005/12/24 01:29:15  Steve
	Added Volume control for 16 bit audio
	Added option to release filter at end of each track or when stopped
	Added tool tips to config dialog
	Fixed Filter selection to use friendly names rather than device paths
	
	Revision 1.4  2005/12/10 18:59:36  Steve
	Fixed crash on pause
	Fixed pause interference with start of new track
	Fixed problem with next track not playing when WAV files end
	
================================================================== */

#include "kssample.h"
#include "out.h"
#include "resource.h"
#include "stdarg.h"
#include "error.h"
#include "tooltips.h"
#include "config.h"
#include "deviceinfo.h"

#define PI_VER2 "v2.4"

#define PI_VER PI_VER2 " (x86)"

#define Zero(s)                     ZeroMemory(&s, sizeof(s))

#define PACKETSIZE (8192)
#define MAX_RES_CHANNELS (2)		// max number of channels supported by the resampler


static int srate, numchan, bps;
volatile int writtentime, w_offset;
extern Out_Module out;
static int last_pause=0;
static int settings_changed = FALSE;
static WAVEFORMATEXTENSIBLE wfx_cur;
static int plugin_volume = 0;
static TTItem *ttitem = NULL;

static int src_samplerate_cur = 0;					// current sample rate of source data
static short *resample_buffer = NULL;				// buffer for resampled sample data
static int resample_last[MAX_RES_CHANNELS];		// last sample of last batch for each channel
static DeviceInfoItem *device_info_list = NULL;	// start of device info list (used to populate output device list box)
// configuration information
static Config *config;

// Kernel Streaming stuff

// this defines how many data packets we will use to stream data to the pin
#define NUMPACKETS	10	// probably around 130.6122449 ms latency (depends on amount given to Write())

struct DATA_PACKET
{
    KSSTREAM_HEADER Header;
    OVERLAPPED      Signal;
};

HRESULT				hr				= S_OK;
PBYTE   				pBuffer		= NULL;
ULONG 				cbBuffer		= 0;
HANDLE 				hStopEvent	= NULL;
CKsAudRenFilter*	pFilter		= NULL;
CKsAudRenPin*		pPin			= NULL;
static int			initialised = FALSE;	// if TRUE, all initialisation at beginning of playback completed successfully

volatile KSSTATE State = KSSTATE_STOP;

// data packet stuff
HANDLE      hEventPool[NUMPACKETS + 1];
DATA_PACKET Packets[NUMPACKETS];
ULONG       cbPartialBuffer = 0;
static CKsEnumFilters* the_enumerator = NULL;



BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH) DisableThreadLibraryCalls((HMODULE)hInst);
	return TRUE;
}



//
// SJM Note:
// despite the misleading title, this actually creates all the filters, so deleting it while
// playing would be kinda bad
//

static CKsEnumFilters *OpenEnumerator(void)
{
	if (!the_enumerator)
	{

		// enumerate audio renderers
	   the_enumerator = new CKsEnumFilters(&hr);

		if(the_enumerator == NULL)
		{
			MessageBox(out.hMainWindow, "Failed to allocate CKsEnumFilters", "Error!", MB_OK);
			return NULL;
		}
		else if(!SUCCEEDED(the_enumerator))
		{
			MessageBox(out.hMainWindow, "Failed to create CKsEnumFilters", "Error!", MB_OK);
			return NULL;
		}

	    GUID  aguidEnumCats[] = { STATIC_KSCATEGORY_AUDIO, STATIC_KSCATEGORY_RENDER };

	    hr = 
	        the_enumerator->EnumFilters
	        (
	            eAudRen,            // create audio render filters ...
	            aguidEnumCats,      // ... of these categories
	            2,                  // There are 2 categories
	            TRUE,               // While you're at it, enumerate the pins
	            FALSE,              // ... but don't bother with nodes
	            TRUE                // Instantiate the filters
	        );

	    if(!SUCCEEDED(hr))
		{
			MessageBox(out.hMainWindow, "CKsEnumFilters::EnumFilters failed", "Error!", MB_OK);
			return NULL;
		}
	}
	return the_enumerator;
}





static void CloseEnumerator(void)
{
	if (the_enumerator) {
		delete the_enumerator;
		the_enumerator = NULL;
	}
}











//
// Chooses which filter to use by comparing its device path to the name held in the config
// if it doesn't get a match (e.g. hardware change or first time) it just picks the first one
//

static CKsAudRenFilter *PickFilter(void)
{
	CKsAudRenFilter	*filter;
	CKsEnumFilters		*pEnumerator;

	pEnumerator = the_enumerator;

	if (!pEnumerator)
		return NULL;


	pEnumerator->m_listFilters.GetHead((CKsFilter**)&filter);	// just grab the first one we find
	{
		LISTPOS lp = pEnumerator->m_listFilters.GetHeadPosition();
		filter = NULL;

		// Pick output device by matching name with current selection
		do {
			pEnumerator->m_listFilters.GetNext(lp, (CKsFilter**)&filter);	// Wave1
			if (filter) {
				if (0 == stricmp(config->out_name,filter->GetName()))
					break;
				filter = NULL;
			}
			else break;
		} while(1);

		if (!filter) {	// no match, pick first filter
			lp = pEnumerator->m_listFilters.GetHeadPosition();
			pEnumerator->m_listFilters.GetNext(lp, (CKsFilter**)&filter);	// Wave1
		}
	}

	return filter;
}




//
// The Configuration dialog
//

static BOOL CALLBACK ConfigDialog(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	
	switch(msg)
	{
	case WM_CREATE:
		return TRUE;

	case WM_DESTROY:
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_OUTPUTLIST));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_ENABLEVOLUMECONTROL));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_RELEASEAUDIODEVICEWHENSTOPPED));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_UPSAMPLE));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_DEFAULTS));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDOK));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDCANCEL));

		return FALSE;

	case WM_SHOWWINDOW:

		{
			// setup tooltips
			ttitem->AddTool(GetDlgItem(hwnd,IDC_OUTPUTLIST),					     	"Choose the output device you wish to stream to");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_ENABLEVOLUMECONTROL),		     	"If checked the WinAmp volume control will work although, if outputting to SPDIF having the volume set to any value other than 100% will cause the bitstream to be modified");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_RELEASEAUDIODEVICEWHENSTOPPED), 	"If unchecked, WinAmp will keep hold of the output device even when stopped, this will prevent other applications from outputting any audio.");// This will also make it impossible to switch to another output plugin without first quitting WinAmp before attempting to play audio through the new plugin.");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_UPSAMPLE), 						     	"If checked, the plugin will oversample the audio passing through it (up to 48KHz) to twice its sample rate, this may increase sound quality depending on the attached hardware");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_DEFAULTS), 						     	"Restores the default configuration");
			ttitem->AddTool(GetDlgItem(hwnd,IDOK),					     					"Accepts and applies the new settings");
			ttitem->AddTool(GetDlgItem(hwnd,IDCANCEL),			     					"Disregards the new settings");
		}

//		{
//			CKsEnumFilters *enumerator;
//			CKsAudRenFilter* filter;
//			int ix;
//
//			enumerator = the_enumerator;//OpenEnumerator();
//			if (!enumerator)
//				return FALSE;
//
//			LISTPOS lp = enumerator->m_listFilters.GetHeadPosition();
//		
//			do {
//				filter = NULL;
//				enumerator->m_listFilters.GetNext(lp, (CKsFilter**)&filter);						// get next entry from list
//				if (!filter) break;																				// quit when no more
//
//				ix = SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_ADDSTRING, 0, (LPARAM)filter->GetFriendlyName());	// add name to list box
//				if (!config->out_name[0] && !ix)
//					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight first entry if no current setup
//				else if (0 == strcmp(filter->GetName(),config->out_name))
//					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight if current
//				SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETITEMDATA, ix, (LPARAM)filter->GetName());					// add device path to list box
//
//			} while(1);
//		}

		{
			DeviceInfoItem *i;
			int ix;

			i = device_info_list;

			while(i) {
				ix = SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_ADDSTRING, 0, (LPARAM)i->GetFriendlyName());			// add name to list box
				if (!config->out_name[0] && !ix)
					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight first entry if no current setup
				else if (0 == strcmp(i->GetDevicePath(),config->out_name))
					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight if current
				SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETITEMDATA, ix, (LPARAM)i->GetDevicePath());				// add device path to list box
				i = i->Next();
			}
		}


//		SendDlgItemMessage(hwnd,IDC_OUTPUTLIST, LB_SETHORIZONTALEXTENT,1000,0);
		{
			int sw,sh;
			RECT r;

			sw = GetSystemMetrics(SM_CXSCREEN);
			sh = GetSystemMetrics(SM_CYSCREEN);
			if (GetWindowRect(hwnd,&r)) {
				SetWindowPos(hwnd,HWND_TOP,(sw-(r.right-r.left))>>1,(sh-(r.bottom-r.top))>>1,0,0,SWP_NOSIZE);
			}
		}

		SendDlgItemMessage(hwnd,IDC_ENABLEVOLUMECONTROL,BM_SETCHECK,(WPARAM)           (config->volume_enabled) ? 1:0, 0);
		SendDlgItemMessage(hwnd,IDC_RELEASEAUDIODEVICEWHENSTOPPED,BM_SETCHECK,(WPARAM) (config->clean_up_after_track)  ? 1:0, 0);
		SendDlgItemMessage(hwnd,IDC_UPSAMPLE,BM_SETCHECK,(WPARAM)                      (config->upsample)              ? 1:0, 0);

		return TRUE;

	case WM_COMMAND:
		switch(LOWORD(wparam))
		{
		case IDCANCEL:
			EndDialog(hwnd,0);
			return TRUE;

		case IDOK:
			{
				int ix;
				ix = SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_GETCURSEL,0,0);
				if (ix != LB_ERR) {
					char *dev_path;
					dev_path = (char *)SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_GETITEMDATA,ix,0);						// retrieve output name
					strcpy(config->out_name,dev_path);
				}

				config->volume_enabled = SendDlgItemMessage(hwnd,IDC_ENABLEVOLUMECONTROL,BM_GETSTATE,0,0) & BST_CHECKED;
				config->clean_up_after_track  = SendDlgItemMessage(hwnd,IDC_RELEASEAUDIODEVICEWHENSTOPPED,BM_GETSTATE,0,0) & BST_CHECKED;
				config->upsample              = SendDlgItemMessage(hwnd,IDC_UPSAMPLE,BM_GETSTATE,0,0) & BST_CHECKED;

//				config_write(out_name,plugin_volume_enabled,clean_up_after_track,upsample);										// update config info
  				config->Write();

				settings_changed = TRUE;
			}
			EndDialog(hwnd,0);
			return (TRUE);

		case IDC_DEFAULTS:
			config->Defaults();
			config->Write();
			settings_changed = TRUE;
			EndDialog(hwnd,0);
			return TRUE;

		default:
			return FALSE;
		}
	default:
		return FALSE;
	}
}


//
// Function to launch the configuration dialog
//

static void BeginConfigDialog(HWND hwnd)
{
	DialogBox(out.hDllInstance,MAKEINTRESOURCE(IDD_CONFIG),hwnd,ConfigDialog);
}




static void About(HWND hwnd)
{
	MessageBox(hwnd,"Chun-Yu's Kernel Streaming Plug-in " PI_VER "\n"
					"Copyright (c) 2002 Chun-Yu Shei.\n\n"
					"Rewritten to improve stability and add extra functionality\n"
					"by Steve Monks 2005\n\n"
					"Compiled on " __DATE__ "\n","About",MB_OK);
}



//
// clean up playback stuff - called when either Winamp closes or output format changes, or at end of each track
// if cleanup_at_end is enabled
//

static void CloseOutput(void)
{
	// wait for all packets to have finished - kind of important before we start deleting stuff
	WaitForMultipleObjects(NUMPACKETS, hEventPool, TRUE, INFINITE);

	// close the ouput pin
	if (pPin) {
	   pPin->ClosePin();
		pPin = NULL;
	}

	// delete the output buffer(s)
	if (pBuffer) {
		delete [] pBuffer;
		pBuffer = NULL;
	}

	// release per buffer resources
	for(int i = 0; i < NUMPACKETS; i++)
	{
		if (hEventPool[i])
			CloseHandle(hEventPool[i]);
		hEventPool[i] = NULL;
	}

	// delete all filters
	CloseEnumerator();

	// mark as uninitialised
	initialised = FALSE;
}



//
// Chooses a suitable filter and creates a pin, events, buffers and anything else that is required
//

static int OpenOutput(WAVEFORMATEXTENSIBLE &wfx)
{

	// Clean up various playback parameters
	CloseOutput();

	OpenEnumerator();			// create filter list

	pFilter = PickFilter();	// pick filter to render to

	if(!pFilter)
	{
		MessageBox(out.hMainWindow, "No filters available for rendering", "Error!", MB_OK);
		return FALSE;
	}

	pPin = pFilter->CreateRenderPin(&wfx.Format, FALSE);

	if (!pPin)
	{
	    // driver can't handle WAVEFORMATEXTENSIBLE, so fall back to
	    // WAVEFORMATEX format descriptor and try again
	    wfx.Format.wFormatTag = WAVE_FORMAT_PCM;
	    // set unused members to zero
	    wfx.Format.cbSize = 0;
	    wfx.Samples.wValidBitsPerSample = 0;
	    wfx.dwChannelMask = 0;
	    wfx.SubFormat = GUID_NULL;

	    pPin = pFilter->CreateRenderPin(&wfx.Format, FALSE);
	}

	if (!pPin)
	{
		MessageBox(out.hMainWindow, "Unable to create output pin", "Error!", MB_OK);
		return FALSE;
	}

	// allocate NUMPACKETS 8192 byte buffers - should *always* work because Write() should never write more than 8192b
	cbBuffer = NUMPACKETS * PACKETSIZE;
	pBuffer = new BYTE[cbBuffer];
	cbPartialBuffer = cbBuffer / NUMPACKETS;
	ZeroMemory(pBuffer, cbBuffer);

	// set up the data packets
	for(int i = 0; i < NUMPACKETS; i++)
	{
		Zero(Packets[i]);
		hEventPool[i] = CreateEvent(NULL, TRUE, TRUE, NULL);	// NO autoreset!
		if (!hEventPool[i])
  		{
  			MessageBox(out.hMainWindow, "Unable to create event", "Error!", MB_OK);
  			return FALSE;
  		}

		Packets[i].Signal.hEvent = hEventPool[i];
		Packets[i].Header.Data = pBuffer + i * cbPartialBuffer;
		Packets[i].Header.FrameExtent = cbPartialBuffer;
		Packets[i].Header.DataUsed = cbPartialBuffer;  // if we were capturing, we would init this to 0
		Packets[i].Header.Size = sizeof(Packets[i].Header);
		Packets[i].Header.PresentationTime.Numerator = 1;
		Packets[i].Header.PresentationTime.Denominator = 1;
	}

	ResetEvent(hStopEvent);	// reset stop event - it isn't even used (yet?)
	hEventPool[NUMPACKETS] = hStopEvent;	// stop event

	wfx_cur = wfx;

	initialised = TRUE;

	return TRUE;
}


//
// Initialise the Plugin
//

static void Initialise()
{
	//MessageBox(out.hMainWindow, "", "Init", MB_OK);
//	config_read(out_name,sizeof(out_name),&plugin_volume_enabled,&clean_up_after_track,&upsample);
	config = new Config();
	hStopEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	OpenEnumerator(); 
	device_info_list = BuildDeviceInfoList(the_enumerator);
	memset(&wfx_cur,0,sizeof(wfx_cur));
	memset(hEventPool,0,sizeof(hEventPool));
	DebugMsgInit(out.hMainWindow);
	plugin_volume = 255;

	resample_buffer = (short *)malloc(PACKETSIZE);
	memset(resample_last,0,sizeof(resample_last));

	ttitem = new TTItem(out.hDllInstance);
}


//
// Close the plugin and clean up
//

static void Quit()
{
	if (ttitem)
		delete ttitem;
	ttitem = NULL;

	if (resample_buffer)
		free(resample_buffer);

	resample_buffer = NULL;

	//MessageBox(out.hMainWindow, "", "Quit", MB_OK);
	CloseOutput();
	DestroyDeviceInfoList();
	CloseEnumerator();
	CloseHandle(hStopEvent);
}









//
// called whenever play is pressed in Winamp or a new track begins playing
//

int StartPlayback(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
//	DebugMsgBox("Open");
	int src_samplerate;

	DebugMsg("Open!");

	src_samplerate = samplerate;
	if (config->upsample && resample_buffer && samplerate <= 48000 && numchannels < MAX_RES_CHANNELS && bitspersamp == 16) {
		samplerate <<= 1;
	}

	// instantiate the pin
    // use WAVEFORMATEXTENSIBLE to describe wave format
	WAVEFORMATEXTENSIBLE wfx;
	wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
	wfx.Format.nChannels = numchannels;
	wfx.Format.nSamplesPerSec = samplerate;
	wfx.Format.nBlockAlign = numchannels*bitspersamp/8;	// I think
	wfx.Format.nAvgBytesPerSec = wfx.Format.nSamplesPerSec * wfx.Format.nBlockAlign;
	wfx.Format.wBitsPerSample = bitspersamp;
	wfx.Format.cbSize = sizeof(WAVEFORMATEXTENSIBLE)-sizeof(WAVEFORMATEX);
	wfx.Samples.wValidBitsPerSample = bitspersamp;
	wfx.dwChannelMask = KSAUDIO_SPEAKER_STEREO;
	wfx.SubFormat = KSDATAFORMAT_SUBTYPE_PCM;


	// format changed or settings changed?
	// delete filters and recreate for new format/settings
	if (memcmp(&wfx,&wfx_cur,sizeof(wfx)) || (src_samplerate != src_samplerate_cur) || settings_changed || !initialised) {
		src_samplerate_cur = src_samplerate;
		if (!OpenOutput(wfx))
			return -1;
	}


	// set new state
	State = KSSTATE_RUN;
	hr = pPin->SetState(State);

	if(!SUCCEEDED(hr))
	{
		MessageBox(out.hMainWindow, "Error in pPin->SetState(KSSTATE_RUN) !", "Error!", MB_OK);
		return -1;
	}

 	w_offset = writtentime = 0;
	numchan = numchannels;
	srate = samplerate;
	bps = bitspersamp;
	last_pause = 0;

	return 0;
}


//
// Called whenever stop is pressed in Winamp or we reach the end of the current track
//

static void EndPlayback()
{

	if (!initialised)
		return;

	// stop the pin
	if (pPin) {
		hr = pPin->SetState(KSSTATE_PAUSE);
		hr = pPin->SetState(KSSTATE_STOP);
	}

	State = KSSTATE_STOP;

	if (config->clean_up_after_track)
		CloseOutput();
}



//
// copies a block of memory with suitable attenuation
//
static void CopyMemoryAttenuated(void *dest, void *source, int len, WAVEFORMATEXTENSIBLE *wfx)
{
	typedef union {
		char *c;
		short *s;
		int *i;
		float *f;
		void *v;
	} VP;

	VP d, s;
	union {
		int i;
		char c[4];
	} v;

	if (plugin_volume != 255 && config->volume_enabled) {
		switch(wfx->Format.wBitsPerSample)
		{
		case 8:
			d.v = dest;
			s.v = source;
			while(len >= 1) {
				v.i = *s.c;
				v.i = (v.i * plugin_volume) >> 8;
				*d.c = v.i;
				++d.c;
				++s.c;
				len -= 1;
			}
			return;

		case 16:
			d.v = dest;
			s.v = source;
			while(len >= 2) {
				v.i = *s.s;
				v.i = (v.i * plugin_volume) >> 8;
				*d.s = v.i;
				++d.s;
				++s.s;
				len -= 2;
			}
			return;

//		case 24:
//			d.v = dest;
//			s.v = source;
//			v.i = 0;
//			while(len >= 3) {
//				v.c[1] = s.c[0];		// read 24 bit sample
//				v.c[2] = s.c[1];
//				v.c[3] = s.c[2];
//				v.i >>= 8;				// so it's sign extended correctly
//				v.i = (v.i * plugin_volume) >> 8;
//				d.c[0] = v.c[0];
//				d.c[1] = v.c[1];
//				d.c[2] = v.c[2];
//				d.c += 3;
//				s.c += 3;
//				len -= 3;
//			}
//			return;

		default:
			// unsupported format, just memory copy it by falling through
			break;
		}	
	}
	// unsupported format, just copy the data unmodified
	CopyMemory(dest,source, len);
}



static int WriteToOutputDevice(char *buf, int len)
{
	// use NUMPACKETS buffers to flip between (later should use more)
	
	// see if a buffer is free
	DWORD dwWait = WaitForMultipleObjects(NUMPACKETS, hEventPool, FALSE, 0);
	// if not, tell input plugin that we aren't ready yet
	if((dwWait == WAIT_FAILED) || (dwWait == WAIT_TIMEOUT)) return 1;
	else dwWait -= WAIT_OBJECT_0;

	CopyMemoryAttenuated(Packets[dwWait].Header.Data, buf, len, &wfx_cur);

	Packets[dwWait].Header.FrameExtent = len;
	Packets[dwWait].Header.DataUsed = len;

	// key - no using autoreset events!!! (yes, I spent 3 hours figuring this out the *hard* way)
	ResetEvent(Packets[dwWait].Signal.hEvent);

	hr = pPin->WriteData(&Packets[dwWait].Header, &Packets[dwWait].Signal);
//	if(!SUCCEEDED(hr)) return 1;	// or does it really matter? SJM:probably not!

	writtentime += len;

	return 0;
}



static void Resample16(short *in, int len)
{
	short *out;
	int n;

	out = resample_buffer;

	len >>= 1;	// convert from bytes to shorts (16 bit samples)

	while(len) {
		for (n = 0; n < wfx_cur.Format.nChannels; n++) {
			out[n] = (resample_last[n] + (int)in[n]) >> 1;
			out[wfx_cur.Format.nChannels] = in[n];
			resample_last[n] = in[n];
		}

		out += wfx_cur.Format.nChannels<<1;
		in  += wfx_cur.Format.nChannels;
		len -= wfx_cur.Format.nChannels;
	}
}



static int WriteBuffer(char *buf, int len)
{
	int len1,len2;

	if (!initialised)
		return 0;

	if (State == KSSTATE_PAUSE)
		return 1;
//		MessageBox(out.hMainWindow, "", "write while paused", MB_OK);

	if (!len)
		return 0;	// this happens at end of file when playing wavs. Oddly enough doesn't happen with .wv's

	if (!pPin)
		return 0;

	if (src_samplerate_cur != wfx_cur.Format.nSamplesPerSec) {	// input sample rate != output sample rate, therefore, upsampling is enabled
		// split the buffer, note, this should be done by the blockalign boundary
		if (len < 4096) {	// complete packet resampled will fit in a single output packet
			len1 = len;
			len2 = 0;
		}
		else {					// will require two packets to send the full thing
			len1 = len >> 1;
			len1 = len1 - (len1 % wfx_cur.Format.nBlockAlign);
			len2 = len - len1;
		}
		Resample16((short *)buf,len1);							// resample first half of buffer
		while(WriteToOutputDevice((char *)resample_buffer,len1<<1)) {
			Sleep(1);
		}

		if (len2) {
			Resample16((short *)(buf+len1),len2);		// resample second half of buffer
			while(WriteToOutputDevice((char *)resample_buffer,len2<<1)) {
				Sleep(1);
			}
		}
		return 0;
	}
	else
		return WriteToOutputDevice(buf,len);
}


static int CanWrite()
{
	if (!initialised)
		return 0;

	if (State == KSSTATE_PAUSE)
		return 0;

	// see if we can write - just "peek" at the event and don't reset it!
	DWORD dwWait = WaitForMultipleObjects(NUMPACKETS, hEventPool, FALSE, 0);
	if(dwWait != WAIT_FAILED && dwWait != WAIT_TIMEOUT) return last_pause ? 0 : PACKETSIZE;
	else return 0;
}

static int IsPlaying()
{
	if (!initialised)
		return 0;

	// check if BOTH (well, later ALL) packets are done
	DWORD result = WaitForMultipleObjects(NUMPACKETS, hEventPool, TRUE, 0);
	if(!(result - WAIT_OBJECT_0))	// I think...
	{
		return 0;	// neither of them timed out, so we must be done playing
	}
	else return 1;	// one of them isn't signaled yet - still playing
}

static int Pause(int pause)
{
	int t=last_pause;
	last_pause=pause;

	if (!initialised)
		return t;

	 Sleep(250);

	if(pause)	// pausing
	{
		if (State == KSSTATE_STOP)	// should *never* happen, but oh, well
		{
			for (int i = 0; i < NUMPACKETS; i++)
				SetEvent(hEventPool[i]);

//			DebugMsgBox("Thing that shouldn't happen - happened");
		}

		// set new state
		State = KSSTATE_PAUSE;
		if (pPin)
			hr = pPin->SetState(State);
	}
	else	// unpausing
	{
		if (State == KSSTATE_STOP)	// should *never* happen, but oh, well
        {
            for (int i = 0; i < NUMPACKETS; i++)
                SetEvent(hEventPool[i]);

//				 DebugMsgBox("Thing that shouldn't happen - happened");
        }

        // set new state
			if (pPin)
   	     hr = pPin->SetState(KSSTATE_RUN);
        State = KSSTATE_RUN;
	}

	 Sleep(250);

	return t;
}

static void SetVolume(int volume)
{
	if (volume >= 0)		// seem to get a negative value passed through immediately after playback begins
		plugin_volume = max(0,min(255,volume));
}

static void SetPan(int pan)
{
	// to be added later
}



static int GetWrittenTime()
{
	// should check output position, but I don't yet
	int t=srate*numchan,l;
	int ms=writtentime;

	l=ms%t;
	ms /= t;
	ms *= 1000;
	ms += (l*1000)/t;

	if (bps == 16) ms/=2;

	return ms + w_offset;
}


static void Flush(int t)
{
	int a;
	w_offset=0;
	a = t - GetWrittenTime();
	w_offset=a;
}
	


Out_Module out = {
	OUT_VER,
	"Kernel Streaming Output "PI_VER2,
	0x10000,	// was 33, but this is the ID of the disk writer, changed, but streaming media still blocks this plugin
	0, // hmainwindow
	0, // hdllinstance
	BeginConfigDialog,
	About,
	Initialise,
	Quit,
	StartPlayback,
	EndPlayback,
	WriteBuffer,
	CanWrite,
	IsPlaying,
	Pause,
	SetVolume,
	SetPan,
	Flush,
	GetWrittenTime,
	GetWrittenTime
};

extern "C"
{
__declspec( dllexport ) Out_Module * winampGetOutModule()
{
	return &out;
}
}
